class AmbiguousLookupError(Exception):
    """
    Raised where a user has configured a lookup engine such that it
    cannot concretely resolve a lookup against one of more observation
    cells.
    """

    def __init__(self, msg):
        self.msg = msg


class MissingDirectLookupError(Exception):
    """
    Raised where a user has specified a direct lookup but there is
    not value in the direction specified.
    """

    def __init__(self, msg):
        self.msg = msg


class FailedLookupError(Exception):
    """
    Raised where a lookup engine fails to resolve a result cell
    """

    def __init__(self, msg):
        self.msg = msg



"""
Helpers to order a list of cells to represent different ways you
could read a tablulated display of data in a cell by cell fashion.
"""

from typing import List

import tidychef.datafuncs as dfc
from tidychef.models.source.cell import BaseCell


def order_cells_leftright_topbottom(cells: List[BaseCell]) -> List[BaseCell]:
    """
    Given a list of BaseCell's sort them into a typical human readable order.

    Example cell read order:
    |  1  |  2  |  3  |
    |  4  |  5  |  6  |

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: A list of BaseCell or inheritors of that represents a selection
    of cells.
    """
    return sorted(cells, key=lambda cell: (cell.y, cell.x), reverse=False)


def order_cells_rightleft_bottomtop(cells: List[BaseCell]) -> List[BaseCell]:
    """
    Given a list of BaseCell's sort them into a reverse human readable order.

    Example cell read order:
    |  6  |  5  |  4  |
    |  3  |  2  |  1  |

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: A list of BaseCell or inheritors of that represents a selection
    of cells.
    """
    return sorted(cells, key=lambda cell: (cell.y, cell.x), reverse=True)


def order_cells_topbottom_leftright(cells: List[BaseCell]) -> List[BaseCell]:
    """
    Order cells moving top to bottom from the leftmost column to the
    rightmost column.

    Example cell read order:
    |  1  |  3  |  5  |
    |  2  |  4  |  6  |

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: A list of BaseCell or inheritors of that represents a selection
    of cells.
    """
    return sorted(cells, key=lambda cell: (cell.x, cell.y), reverse=False)


def order_cells_bottomtop_rightleft(cells: List[BaseCell]) -> List[BaseCell]:
    """
    Order cells moving bottom to top from the rightmost column to the
    leftmost column.

    Example cell read order:
    |  6  |  4  |  2  |
    |  5  |  3  |  1  |

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: A list of BaseCell or inheritors of that represents a selection
    of cells.
    """
    return sorted(cells, key=lambda cell: (cell.x, cell.y), reverse=True)


def order_cells_rightleft_topbottom(cells: List[BaseCell]) -> List[BaseCell]:
    """
    Order cells left to right moving from the bottom row to the
    top row.

    Example cell read order:
    |  3  |  2  |  1  |
    |  6  |  5  |  4  |

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: A list of BaseCell or inheritors of that represents a selection
    of cells.
    """
    maximum_y = max(cell.y for cell in cells)
    maximum_x = max(cell.x for cell in cells)

    ordered_cells = []
    for y_index in range(maximum_y + 1, -1, -1):
        for x_index in range(maximum_x + 1, -1, -1):
            wanted_cell = BaseCell(x=x_index, y=y_index)
            if dfc.cell_is_within(cells, wanted_cell):
                ordered_cells.append(
                    dfc.specific_cell_from_xy(cells, x_index=x_index, y_index=y_index)
                )
    return ordered_cells


def order_cells_topbottom_rightleft(cells: List[BaseCell]) -> List[BaseCell]:
    """
    Order cells right to left moving from the lefthand to right columns
    considering cells top to bottom

    Example cell read order:
    |  5  |  3  |  1  |
    |  6  |  4  |  2  |

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: A list of BaseCell or inheritors of that represents a selection
    of cells.
    """
    maximum_y = max(cell.y for cell in cells)
    maximum_x = max(cell.x for cell in cells)

    ordered_cells = []
    for y_index in range(0, maximum_y + 1):
        for x_index in range(maximum_x + 1, -1, -1):
            wanted_cell = BaseCell(x=x_index, y=y_index)
            if dfc.cell_is_within(cells, wanted_cell):
                ordered_cells.append(
                    dfc.specific_cell_from_xy(cells, x_index=x_index, y_index=y_index)
                )
    return ordered_cells


def order_cells_leftright_bottomtop(cells: List[BaseCell]) -> List[BaseCell]:
    """
    Order cells left to right moving from the left to righthand columns
    considering cells bottom to top

    Example cell read order:
    |  2  |  4  |  6  |
    |  1  |  3  |  5  |

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: A list of BaseCell or inheritors of that represents a selection
    of cells.
    """
    maximum_y = max(cell.y for cell in cells)
    maximum_x = max(cell.x for cell in cells)

    ordered_cells = []
    for x_index in range(0, maximum_x + 1):
        for y_index in range(maximum_y + 1, -1, -1):
            wanted_cell = BaseCell(x=x_index, y=y_index)
            if dfc.cell_is_within(cells, wanted_cell):
                ordered_cells.append(
                    dfc.specific_cell_from_xy(cells, x_index=x_index, y_index=y_index)
                )
    return ordered_cells


def order_cells_bottomtop_leftright(cells: List[BaseCell]) -> List[BaseCell]:
    """
    Order cells from bottom row to top row moving from left to right

    Example cell read order:
    |  4  |  5  |  6  |
    |  1  |  2  |  3  |

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: A list of BaseCell or inheritors of that represents a selection
    of cells.
    """
    maximum_y = max(cell.y for cell in cells)
    maximum_x = max(cell.x for cell in cells)

    ordered_cells = []
    for y_index in range(maximum_y + 1, -1, -1):
        for x_index in range(0, maximum_x + 1):
            wanted_cell = BaseCell(x=x_index, y=y_index)
            if dfc.cell_is_within(cells, wanted_cell):
                ordered_cells.append(
                    dfc.specific_cell_from_xy(cells, x_index=x_index, y_index=y_index)
                )
    return ordered_cells



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tidychef',
 'tidychef.acquire',
 'tidychef.acquire.csv',
 'tidychef.acquire.python',
 'tidychef.acquire.xls',
 'tidychef.acquire.xlsx',
 'tidychef.against',
 'tidychef.against.implementations',
 'tidychef.column',
 'tidychef.datafuncs',
 'tidychef.direction',
 'tidychef.exceptions',
 'tidychef.lookup',
 'tidychef.lookup.engines',
 'tidychef.models',
 'tidychef.models.source',
 'tidychef.notebook',
 'tidychef.notebook.preview',
 'tidychef.notebook.preview.html',
 'tidychef.output',
 'tidychef.selection',
 'tidychef.selection.csv',
 'tidychef.selection.filters',
 'tidychef.selection.xls',
 'tidychef.selection.xlsx',
 'tidychef.utils',
 'tidychef.utils.cellutils',
 'tidychef.utils.decorators',
 'tidychef.utils.fileutils',
 'tidychef.utils.http']

package_data = \
{'': ['*']}

install_requires = \
['cachecontrol>=0.13.1,<0.14.0',
 'filelock>=3.12.2,<4.0.0',
 'jupyter>=1.0.0,<2.0.0',
 'openpyxl>=3.1.2,<4.0.0',
 'requests>=2.31.0,<3.0.0',
 'tabulate>=0.9.0,<0.10.0',
 'validators>=0.20.0,<0.21.0',
 'xlrd>=2.0.1,<3.0.0']

setup_kwargs = {
    'name': 'tidychef',
    'version': '0.1.0',
    'description': 'Python package for transforming human readable tables into tidy data',
    'long_description': 'None',
    'author': 'mikeAdamss',
    'author_email': 'mikelovesbooks@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



BORDER_CELL_COLOUR = "lightgrey"
BORDER_CELL_SECONDARY_COLOUR = "#999999"
NO_COLOUR = "white"
WARNING_COLOUR = "#ff8080"

COLOURS = [
    "cyan",
    "#99ff99",
    "#eeccff",
    "#ffe066",
    "#ff4da6",
    "#ff9933",
    "#4d4dff",
    "#b3d9ff",
    "#00b3b3",
    "#99ffcc",
    "#b380ff",
]

# Simple CSS to make it pretty-ish
INLINE_CSS = """
    <style>
    table, th, td {
        border: 1px solid;
    }

    table {
        border-collapse: collapse;
    }

    td {
        align: center;
        border: 1px  black solid !important;
        color: black !important;
    }

    th, td {
        padding: 5px;
    }

    </style>
    """



"""
Data functions related to excel referencing of x.y positioned cells.
"""
import re
from typing import List

import tidychef.datafuncs.common as dfccommon
from tidychef.exceptions import BadExcelReferenceError, ReversedExcelRefError
from tidychef.models.source.cell import BaseCell
from tidychef.utils import cellutils


def assert_excel_ref_within_cells(cells: List[BaseCell], excel_ref: str):
    """
    Given a list of cells, assert that the cell denoted by
    the provided excel reference, exists within the list.

    :param cells: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :param excel_ref: A single excel reference for a single cell.
    """
    wanted_cell: BaseCell = single_excel_ref_to_basecell(excel_ref)
    match = [c1 for c1 in cells if any([c1.matches_xy(wanted_cell)])]
    assert (
        len(match) == 1
    ), f"Cell of excel ref {excel_ref} is not contained in the provided selection"


def any_excel_ref_as_wanted_basecells(excel_ref: str) -> List[BaseCell]:
    """
    Convert an excel reference of either the single or multiple cell
    format into a list of BaseCell objects.

    :param excel_ref: A excel style reference to one or more cells.
    :return: A list of BaseCell or inheritors of that represents a selection
    of cells.
    """

    # Single style reference, eg: A6, ZA18 etc
    if re.match("^[A-Z]+[0-9]+$", excel_ref):
        return [single_excel_ref_to_basecell(excel_ref)]

    # Multi style reference, eg: A1:Z78, G15:AV16 etc
    if re.match("^[A-Z]+[0-9]+:[A-Z]+[0-9]+$", excel_ref):
        return multi_excel_ref_to_basecells(excel_ref)

    raise BadExcelReferenceError(
        f"Could not identify style of excel reference {excel_ref}"
    )


def single_excel_row_to_y_index(excel_ref: str) -> int:
    """
    Given a single row number as an excel reference, return
    the y offset value

    :param excel_ref: A number representing an excel style
    reference to a single row in a tabulated data source.
    :return: y (vertical offset) value.
    """
    return cellutils.number_to_y(int(excel_ref))


def single_excel_column_to_x_index(excel_ref: str) -> str:
    """
    Given a single column latter as an excel reference, return
    the x index.

    :param excel_ref: One or more letters representing an excel
    style reference to a single column in a tabulated source.
    :return: x (horizontal offset) value.
    """
    return cellutils.letters_to_x(excel_ref)


def single_excel_ref_to_basecell(excel_ref: str) -> BaseCell:
    """
    Given a single excel cell reference, return a single BaseCell
    or inheritor of.

    :param excel_ref: A excel style reference to a single cell.
    :return: A BaseCell or inheritor of.
    """

    letters = ""
    number = None
    for i, letter_or_num in enumerate(excel_ref):
        try:
            int(letter_or_num)
            number = int(excel_ref[i:])
            break
        except ValueError:
            letters += letter_or_num

    assert number
    assert len(letters) > 0

    x_index = cellutils.letters_to_x(letters)
    y_index = cellutils.number_to_y(number)

    return BaseCell(x=x_index, y=y_index)


def multi_excel_ref_to_basecells(excel_ref: str) -> List[BaseCell]:
    """
    Given an excel reference referring to multiple cells, return a list of
    wanted BaseCells.

    :param excel_ref: A excel style reference to a range of cells.
    :return: A list of BaseCell or inheritors of that represents a selection
    of cells.
    """

    assert (
        ":" in excel_ref
    ), """
        A multi cell excel reference should be in the format:
        <letter(s)><number>:<letter(s)><number>

        {excel_ref} is missing a colon, ":"
    """
    start_cell = single_excel_ref_to_basecell(excel_ref.split(":")[0])
    end_cell = single_excel_ref_to_basecell(excel_ref.split(":")[1])

    if start_cell.x > end_cell.x or start_cell.y > end_cell.y:
        raise ReversedExcelRefError(
            """
            Invalid excel reference format. Please provide your reference in the
            standard upmost left to downmost right format, i.e A1:C5 not C5:A1
                """
        )

    start_x = start_cell.x
    start_y = start_cell.y
    end_x = end_cell.x
    end_y = end_cell.y

    return_cells = []
    for x_index in range(start_x, end_x + 1):
        for y_index in range(start_y, end_y + 1):
            return_cells.append(BaseCell(x=x_index, y=y_index))

    return return_cells


def basecell_to_excel_ref(cell: BaseCell) -> str:
    """
    Given a single BaseCell object, return the representative excel reference.

    :param: A single BaseCell or inheritor of
    :return: A excel style reference to a single cell.
    """
    return f"{cellutils.x_to_letters(cell.x)}{cellutils.y_to_number(cell.y)}"


def basecells_to_excel_ref(cells: List[BaseCell]) -> str:
    """
    Given a list of cells representing a solid selection with
    no gaps. Return the representative excel reference.

    :param: A list of BaseCell or inheritors of that represents a selection
    of cells.
    :return: A excel style reference to multiple cells.
    """

    min_x, max_x, min_y, max_y = dfccommon.assert_quadrilaterals(cells)

    topleft_cell = dfccommon.specific_cell_from_xy(cells, min_x, min_y)
    bottomright_cell = dfccommon.specific_cell_from_xy(cells, max_x, max_y)

    ref1 = f"{cellutils.x_to_letters(topleft_cell.x)}{cellutils.y_to_number(topleft_cell.y)}"
    ref2 = (
        f"{cellutils.x_to_letters(bottomright_cell.x)}"
        f"{cellutils.y_to_number(bottomright_cell.y)}"
    )
    return f"{ref1}:{ref2}"



"""
Holds the code that defines the local xlsx reader.
"""

import io
from pathlib import Path
from typing import Callable, List, Optional, Union

import openpyxl
import requests
import validators

from tidychef.acquire.base import BaseReader
from tidychef.models.source.cell import Cell
from tidychef.models.source.table import Table
from tidychef.selection.selectable import Selectable
from tidychef.selection.xlsx.xlsx import XlsxSelectable
from tidychef.utils.http.caching import get_cached_session

from ..base import BaseReader
from ..main import acquirer


def http(
    source: Union[str, Path],
    selectable: Selectable = XlsxSelectable,
    pre_hook: Optional[Callable] = None,
    post_hook: Optional[Callable] = None,
    session: requests.Session = None,
    cache: bool = True,
    tables: str = None,
    **kwargs,
) -> Union[XlsxSelectable, List[XlsxSelectable]]:
    """
    Read data from a Path (or string representing a path)
    present on the same machine where tidychef is running.

    This local xlsx reader uses openpyxl:
    https://openpyxl.readthedocs.io/en/stable/index.html

    Any kwargs passed to this function are propagated to
    the openpyxl.load_workbook() method.

    :param source: A url.
    :param selectable: A class that implements tidychef.selection.selectable.Selectable of an inheritor of. Default is XlsxSelectable
    :param pre_hook: A callable that can take source as an argument
    :param post_hook: A callable that can take the output of XlsxSelectable.parse() as an argument.
    :param session: An optional requests.Session object.
    :param cache: Boolean flag for whether or not to cache get requests.
    :return: A single populated Selectable of type as specified by selectable param.
    """

    assert validators.url(source), f"'{source}' is not a valid http/https url."

    return acquirer(
        source,
        HttpXlsxReader(tables),
        selectable,
        pre_hook=pre_hook,
        post_hook=post_hook,
        session=session,
        cache=cache,
        **kwargs,
    )


class HttpXlsxReader(BaseReader):
    """
    A reader to lead in a source where that source is a locally
    held xlsx file.
    """

    def parse(
        self,
        source: str,
        selectable: Selectable = XlsxSelectable,
        data_only=True,
        session: requests.Session = None,
        cache: bool = True,
        **kwargs,
    ) -> List[XlsxSelectable]:
        """
        Parse the provided source into a list of Selectables. Unless overridden the
        selectable is of type XlsxSelectable.

        Additional **kwargs are propagated to openpyxl.load_workbook()

        :param source: A url.
        :param selectable: The selectable type to be returned.
        :data_only: An openpyxl.load_workbook() option to disable acquisition of non data elements from the tabulated source (macros etc)
        :param session: An optional requests.Session object.
        :param cache: Boolean flag for whether or not to cache get requests.
        :return: A list of type as specified by param selectable.
        """

        if not session:
            if cache:
                session = get_cached_session()
            else:
                session = requests.session()

        response: requests.Response = session.get(source, **kwargs)
        if not response.ok:
            raise requests.exceptions.HTTPError(
                f"""
                Unable to get url: {source}
                {response}
                """
            )

        bio = io.BytesIO()
        bio.write(response.content)
        bio.seek(0)

        workbook: openpyxl.Workbook = openpyxl.load_workbook(
            bio, data_only=data_only, **kwargs
        )

        tidychef_selectables = []
        worksheet_names = workbook.get_sheet_names()
        for worksheet_name in worksheet_names:

            worksheet = workbook.get_sheet_by_name(worksheet_name)

            table = Table()
            for y, row in enumerate(worksheet.iter_rows()):
                for x, cell in enumerate(row):
                    table.add_cell(
                        Cell(x=x, y=y, value=str(cell.value) if cell.value else "")
                    )

            tidychef_selectables.append(
                selectable(table, source=source, name=worksheet_name)
            )
        return tidychef_selectables



"""
Holds the code that defines the local xlsx reader.
"""

from pathlib import Path
from typing import Callable, List, Optional, Union

import xlrd

from tidychef.acquire.base import BaseReader
from tidychef.models.source.cell import Cell
from tidychef.models.source.table import Table
from tidychef.selection.selectable import Selectable
from tidychef.selection.xls.xls import XlsSelectable

from ..base import BaseReader
from ..main import acquirer


def local(
    source: Union[str, Path],
    selectable: Selectable = XlsSelectable,
    pre_hook: Optional[Callable] = None,
    post_hook: Optional[Callable] = None,
    tables: str = None,
    **kwargs,
) -> Union[XlsSelectable, List[XlsSelectable]]:
    """
    Read data from a Path (or string representing a path)
    present on the same machine where tidychef is running.

    This xls reader uses xlrd:
    https://xlrd.readthedocs.io/en/latest/

    Any kwargs passed to this function are propagated to
    the xlrd.open_workbook() method.

    :param source: A Path object or a string representing a path
    :param selectable: A class that implements tidychef.selection.selectable.Selectable of an inheritor of. Default is XlsSelectable
    :param pre_hook: A callable that can take source as an argument
    :param post_hook: A callable that can take the output of XlsSelectable.parse() as an argument.
    :return: A single populated Selectable of type as specified by selectable param
    """

    assert isinstance(
        source, (str, Path)
    ), """
        The source you're passing to acquire.csv.local() needs to
        be either a Path object or a string representing such.
        """

    return acquirer(
        source,
        LocalXlsReader(tables),
        selectable,
        pre_hook=pre_hook,
        post_hook=post_hook,
        **kwargs,
    )


class LocalXlsReader(BaseReader):
    """
    A reader to lead in a source where that source is a locally
    held xls file.
    """

    def parse(
        self,
        source: Union[str, Path],
        selectable: Selectable = XlsSelectable,
        **kwargs,
    ) -> List[XlsSelectable]:
        """
        Parse the provided source into a list of Selectables. Unless overridden the
        selectable is of type XlsSelectable.

        Additional **kwargs are propagated to xlrd.open_workbook()

        :param source: A Path or str representing a path indicating a local file
        :param selectable: The selectable type to be returned.
        :return: A list of type as specified by param selectable.
        """

        workbook: xlrd.Book = xlrd.open_workbook(source)
        assert isinstance(workbook, xlrd.Book)

        tidychef_selectables = []
        worksheet_names = workbook.sheet_names()
        for worksheet_name in worksheet_names:

            worksheet = workbook.sheet_by_name(worksheet_name, **kwargs)

            table = Table()
            num_rows = worksheet.nrows
            for y in range(0, num_rows):
                for x, cell in enumerate(worksheet.row(y)):
                    table.add_cell(
                        Cell(x=x, y=y, value=str(cell.value) if cell.value else "")
                    )

            tidychef_selectables.append(
                selectable(table, source=source, name=worksheet_name)
            )
        return tidychef_selectables



from abc import ABCMeta, abstractmethod

from tidychef.models.source.cell import Cell


class BaseLookupEngine(metaclass=ABCMeta):
    """
    The base class all lookup engines are built upon.
    """

    @abstractmethod
    def resolve(self, cell: Cell, *args) -> Cell:
        """
        Given a single observation cell, resolve
        the lookup, returning the relevant cell
        value as defined by this visual relationship.
        """
